# coding=utf-8
import pandas as pd


def search_api(query, client, date_column=[]):
    """
    Performs a query in the platform and returns the results as a dataframe [1]_.

    Parameters
    ----------
    query : json
        MQL query in JSON format
    client : SmartObjectsClient
        An instance of the SmartObjectsClient
    date_column : array_like, optional
        A list of columns to be transformed from string to datetime

    Returns
    -------
    results_dataframe : dataframe
        A dataframe with the query results

    References
    ----------
    .. [1] mnubo SmartObjects Python client
       https://github.com/mnubo/smartobjects-python-client

    Examples
    --------
    >>> from smartobjects import SmartObjectsClient
    >>> hostname = 'https://rest.sandbox.mnubo.com'
    >>> client_id = "..."
    >>> client_secret = "..."
    >>> mnubo_client = SmartObjectsClient(client_id, client_secret, hostname)
    >>> query = {"from": "event","select": [{"count": "*"}]}
    >>> df_query = mdspy.smartobjects.search_api(query, mnubo_client)
        COUNT(*)
    0  126056211
    """

    response = client.search.search(query).raw
    df_search = pd.io.json.json_normalize(response, 'rows')
    if len(response['rows']) == 0:
        return []
    df_search.columns = [col['label'] for col in response['columns']]
    for column in date_column:
        df_search[column] = pd.to_datetime(df_search[column])
    return df_search



# -*- coding: utf-8 -*-

from .context import mdspy
import unittest
import pandas as pd

reload(mdspy.time_series)


class TTTTestSuite(unittest.TestCase):
    """Basic test cases."""

    def test_ttt_ratio(self):
        df_test = pd.Series(
            index=pd.to_datetime(['2017-01-01', '2017-01-02', '2017-01-03', '2017-01-04', '2017-01-05']),
            data=[10, 5, 10, 5, 4])
        target = 0
        self.assertAlmostEqual(2.66666666, mdspy.ttt.ttt_ratio(df_test, target))

    def test_ttt_ratio_ma(self):
        df_test = pd.Series(
            index=pd.to_datetime(['2017-01-01', '2017-01-02', '2017-01-03', '2017-01-04', '2017-01-05']),
            data=[10, 5, 10, 5, 4])
        target = 0
        self.assertAlmostEqual(5.3333333, mdspy.ttt.ttt_ratio_ma(df_test, target, 2))

    def test_ttt_linear_regression(self):
        df_test = pd.Series(
            index=pd.to_datetime(['2017-01-01', '2017-01-02', '2017-01-03', '2017-01-04', '2017-01-05']),
            data=[10, 5, 10, 5, 4])
        target = 0
        self.assertAlmostEqual(3.3333333, mdspy.ttt.ttt_linear_regression(df_test, target))


if __name__ == '__main__':
    unittest.main()



from statsmodels.tsa.seasonal import seasonal_decompose

def decompose(series, model='additive', filt=None, freq=None):
    """
        Wrapper of the `seasonal_decompose` function of the statsmodels package. Quoting the statsmodels documentation,
        this function decomposes a time series using moving averages

        Parameters
        ----------
        series : array_like
            Input time series
        model : str {"additive", "multiplicative"}
            Type of seasonal component. Abbreviations are accepted.
        filt : array_like
            The filter coefficients for filtering out the seasonal component.
            The concrete moving average method used in filtering is determined by two_sided.
        freq : int, optional
            Frequency of the series. Must be used if x is not a pandas object.
            Overrides default periodicity of x if x is a pandas object with a timeseries index.

        Returns
        -------
        obj
            A object with seasonal, trend, and resid attributes.

        Notes
        -----
        For more details, see the documentation of statsmodels [1]_

        References
        ----------
        .. [1] Function seasonal_decompose() from statsmodels documentation
           http://www.statsmodels.org/dev/generated/statsmodels.tsa.seasonal.seasonal_decompose.html

        Examples
        --------

        >>> import numpy as np
        >>> import matplotlib.pylab as plt
        >>>
        >>> angles = np.arange(1000) * 2 * 10 * np.pi / 1000
        >>> noise = np.random.normal(0, .05, 1000)
        >>> signal = np.sin(angles) + noise
        >>> trend = np.arange(1000.) / 500
        >>> signal = trend + signal
        >>>
        >>> decomp = sm.tsa.seasonal_decompose(signal, freq=200)
        >>> decomp.plot()
        >>> plt.show()
        <matplotlib.figure.Figure at 0x11cfd3a90>
    """
    decomp = seasonal_decompose(series, model=model, filt=filt, freq=freq)
    return decomp



# -*- coding: utf-8 -*-

from .context import mdspy

import unittest

reload(mdspy.time_series)


class TimeSeriesTestSuite(unittest.TestCase):
    """Basic test cases."""

    def test_decompose(self):
        from statsmodels.tsa.seasonal import seasonal_decompose
        import numpy as np

        angles = np.arange(1000) * 2 * 4 * np.pi / 1000
        noise = np.random.normal(0, .05, 1000)
        signal = angles + noise
        trend = np.arange(1000) / 500
        signal = trend + angles

        decomp = mdspy.time_series.decompose(signal, freq=200)
        ref = seasonal_decompose(signal, freq=200)

        np.testing.assert_array_equal(decomp.trend, ref.trend)
        np.testing.assert_array_equal(decomp.seasonal, ref.seasonal)
        np.testing.assert_array_equal(decomp.resid, ref.resid)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from .context import mdspy

import unittest
import numpy as np

reload(mdspy.anomaly_detection)


class AnomalyDetectionTestSuite(unittest.TestCase):
    """Basic test cases."""

    def test_moving_average(self):
        output = {'distance': 17.666918237202548, 'prediction': 0.15, 'range': 0.048112522432468816}
        self.assertEqual(mdspy.anomaly_detection.moving_average([0.15, 0.1, 0.15, 0.2, 0.1], 1., ma=3, h=3), output)

    def test_single_exponential(self):
        output_1 = {'distance': 14.087944, 'prediction': 0.124555, 'range': 0.062141}
        self.assertEqual(
            mdspy.anomaly_detection.single_exponential_smoothing([0.15, 0.1, 0.15, 0.2, 0.1], 1., alpha=0.7, h=3),
            output_1)
        output_2 = {'distance': 12.727922, 'prediction': 0.1, 'range': 0.070711}
        self.assertEqual(
            mdspy.anomaly_detection.single_exponential_smoothing([0.15, 0.1, 0.15, 0.2, 0.1], 1., alpha=1., h=3),
            output_2)
        # test from Section 6.4.3.1 Single Exponential Smoothing, of NIST/SEMATECH e-Handbook of Statistical Methods
        # (see documentation)
        input = [71, 70, 69, 68, 64, 65, 72, 78, 75, 75, 75, 70]
        pred_ref = [np.nan, 71, 70.9, 70.71, 70.44, 69.8, 69.32, 69.58, 70.43, 70.88, 71.29, 71.67]
        pred = [np.nan, 71]
        for i in range(2, 11):
            pred.append(round(mdspy.anomaly_detection.single_exponential_smoothing(input[:i], input[i + 1],
                                                                                   alpha=0.1, h=1)['prediction'], 2))
        pred.append(round(mdspy.anomaly_detection.single_exponential_smoothing(input[:-1], input[-1],
                                                                               alpha=0.1, h=1)['prediction'], 2))
        self.assertEqual(pred_ref, pred)

    def test_double_exponential(self):
        # test from Section 6.4.3.4 Forecasting with Exponential Smoothing, of NIST/SEMATECH e-Handbook of Statistical
        # Methods (see documentation)
        input = np.array([6.4, 5.6, 7.8, 8.8, 11., 11.6, 16.7, 15.3, 21.6, 22.4])
        pred_ref = [np.nan, np.nan, np.nan, np.nan, 9.1, 11.4, 13.2, 17.4, 18.9, 23.1]
        pred = [np.nan, np.nan, np.nan, np.nan]
        for i in range(4, 10):
            pred.append(round(mdspy.anomaly_detection.double_exponential_smoothing(input[:i], input[i],
                                                                                   0.3623, 1., h=1)['prediction'], 1))
        self.assertEqual(pred_ref, pred)

    if __name__ == '__main__':
        unittest.main()



# -*- coding: utf-8 -*-

from .context import mdspy

reload(mdspy)

import unittest
from smartobjects import SmartObjectsClient


class SmartObjectsTestSuite(unittest.TestCase):
    """Basic test cases."""

    def test_search_api(self):
        hostname = 'https://rest.sandbox.mnubo.com'
        client_id = "52WnB4443RqSRtinp1ROw8sG4POKrP5LOfPuXctGQWIblRRZD6"
        client_secret = "MvODHcM48EF0sBds7lA82ZpsfqJOIT2Dxd1dzUh16M7Zk5fnYr"
        mnubo_client = SmartObjectsClient(client_id, client_secret, hostname)

        query = {
            "from": "event",
            "select": [
                {
                    "count": "*"
                }
            ]
        }
        df_query = mdspy.smartobjects.search_api(query, mnubo_client)
        self.assertEqual(df_query['COUNT(*)'].values[0], 126056211)

        query = {
            "from": "event",
            "groupByTime": {
                "field": "x_timestamp",
                "interval": "month"
            },
            "select": [
                {
                    "count": "*"
                }
            ]
        }
        df_query = mdspy.smartobjects.search_api(query, mnubo_client, ['month'])
        self.assertEqual(df_query[df_query['month'] == '2000-01-01']['COUNT(*)'].values[0], 13)

    def test_empty_results(self):
        hostname = 'https://rest.sandbox.mnubo.com'
        client_id = "52WnB4443RqSRtinp1ROw8sG4POKrP5LOfPuXctGQWIblRRZD6"
        client_secret = "MvODHcM48EF0sBds7lA82ZpsfqJOIT2Dxd1dzUh16M7Zk5fnYr"
        mnubo_client = SmartObjectsClient(client_id, client_secret, hostname)

        query = {
            "from": "event",
            "where": {
                "and": [
                    {
                        "x_timestamp": {
                            "dateBetween": "1900-01-01T00:00:00-05:00",
                            "and": "1900-01-02T00:00:00-05:00"
                        }
                    }
                ]
            },
            "select": [
                {
                    "value": "button"
                }
            ]
        }
        df_query = mdspy.smartobjects.search_api(query, mnubo_client)
        print df_query
        self.assertEqual(df_query, [])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from .context import mdspy

import unittest
import pandas as pd
import numpy as np

class AnalyticsTestSuite(unittest.TestCase):
    """Basic test cases."""

    def test_correlation(self):
        t_correlation = mdspy.analytics.correlation([1.0, 2.0, 2.0], [1.0, 2.0, 2.0])
        self.assertEqual(t_correlation['pearson_correlation_coefficient'], 1.0)
        self.assertEqual(t_correlation['interpretation'], 'Strong positive correlation')

    def test_clustering_density(self):
        dataset = [[1, 1], [1, 2], [1, 2], [5, 10], [2, 10]]
        df_clustering = pd.DataFrame(dataset)
        df_clustering.columns = ['col1', 'col2']
        t_clustering = mdspy.analytics.clustering_density(df_clustering, ['col1', 'col2'], 2, 4, 50, 42)
        self.assertEqual(t_clustering['clusters'], [1, 1, 1, 0, 0])
        self.assertEqual(t_clustering['cluster_distribution'], {1: 3, 0: 2})
        self.assertAlmostEqual(t_clustering['centroids'], [[3.5, 10.0], [1.0, 1.6666666666666665]])
        self.assertAlmostEqual(t_clustering['sum_intra_cluster_distances'], 5.1666666666666661)
        self.assertAlmostEqual(t_clustering['sum_inter_cluster_distances'], 75.694444444444443)

    def test_clustering_hierarchical(self):
        dataset = [[1, 1], [1, 2], [1, 2], [5, 10], [2, 10]]
        df_clustering = pd.DataFrame(dataset)
        df_clustering.columns = ['col1', 'col2']
        t_clustering = mdspy.analytics.clustering_hierarchical(df_clustering, ['col1', 'col2'], 2)
        print t_clustering
        self.assertEqual(t_clustering['clusters'], [1, 1, 1, 2, 2])
        self.assertEqual(t_clustering['cluster_distribution'], {1: 3, 2: 2})
        self.assertAlmostEqual(t_clustering['centroids'], [[1.0, 1.6666666666666667], [3.5, 10.0]])
        self.assertAlmostEqual(t_clustering['sum_intra_cluster_distances'], 5.1666666666666661)
        self.assertAlmostEqual(t_clustering['sum_inter_cluster_distances'], 75.694444444444443)

    def test_best_fit_distribution(self):
        x = [1, 1, 1, 2, 1, 2, 5, 10, 20]
        fit_distribution = mdspy.analytics.best_fit_correlation(x)
        self.assertEqual(fit_distribution['distribution'], 'logistic')
        self.assertAlmostEqual(fit_distribution['kolmogorov_smirnof_test'], 0.31093774290280862)
        mu, sigma = 0, 2
        np.random.seed(42)
        x = np.random.normal(mu, sigma, 100)
        fit_distribution = mdspy.analytics.best_fit_correlation(x)
        self.assertEqual(fit_distribution['distribution'], 'dgamma')
        self.assertAlmostEqual(fit_distribution['kolmogorov_smirnof_test'], 0.99576982362294331)


if __name__ == '__main__':
    unittest.main()



import numpy as np


def moving_average(series, sample, ma=20, h=30):
    """
        Predicts the value of the sample based on the average value calculated on the last values of the series.
        Computes the error between prediction and observation and normalizes with the RMSE of errors.

        Parameters
        ----------
        series : array_like
            Input time series
        sample : float
            Sample (observation) to compare the prediction to
        ma : int
            Size of the rolling window (default: 20 samples)
        h : int
            Size of the history (default: 30 samples)

        Returns
        -------
        dict
            `distance`: the distance, expressed in standard deviations of the history values \n
            `prediction`: the prediction for the last value of the input time series \n
            `range`: the standard deviation of of the errors in the history

        Notes
        -----
        The RMSE is calculated on the values of the history.
        The predictions are computed using a moving average long of the "ma" previous samples.
        Because of the normalization, the distance value is to be interpreted in RMSE unit.

        As a rule of thumb, an absolute distance lower than 1 or 2 is acceptable. An absolute value above would probably
        be the sign of an anomaly.

        References
        ----------

        Examples
        --------
        >>> mdspy.anomaly_detection.moving_average([0.15, 0.1, 0.15, 0.2, 0.1], 1., ma=3, h=3)
        {'distance': 17.666918237202548, 'prediction': 0.15, 'range': 0.048112522432468816}
    """

    assert isinstance(series, list)
    result = series[:ma]
    for i in range(ma, len(series)):
        result.append(np.mean(series[i - ma:i]))
    assert len(result) == len(series)
    result.append(np.mean(series[len(series) - ma:]))
    mse = []
    for i in range(len(series) - h, len(series)):
        mse.append(abs(series[i] - result[i]) ** 2)
    rmse = np.sqrt(np.mean(mse))
    output = {'distance': abs(sample - result[-1]) / rmse, 'prediction': result[-1], 'range': rmse}
    return output


def single_exponential_smoothing(series, sample, alpha=0.7, h=10):
    '''
        Predicts the value of the `sample` based on the exponential smoothing algorithm applied to the series of
        previous values.
        The prediction is also used to compute the error with the observation, normalized with the RMSE of previous
        errors.

        Parameters
        ----------
        series : array_like
            Input time series
        sample : float
            Sample (observation) to compare the prediction to
        alpha : float
            Weight (or smoothing factor) to assign to the previous sample
        h : int
            Size of the history (default: 10 samples)

        Returns
        -------
        dict
            `distance`: the distance, expressed in standard deviations of the history values \n
            `prediction`: the prediction for the last value of the input time series \n
            `range`: the standard deviation of of the errors in the history

        Notes
        -----
        The formula for computing the prediction is [1]_

        .. math::
            S_t = a * y_t + (1 - a) * S_{t-1}

        with :math:`S` being the estimate of :math:`y`, :math:`y` the observation and :math:`a` being the smoothing
        factor (or weight). Initialization is :math:`S_0 = y_0`.

        For the forecasting purpose, :math:`S_t` is considered as the prediction
        of the sample :math:`y_{t+1}`, computed with :math:`y_t` and :math:`S_{t-1}`.

        References
        ----------
        .. [1] Section 6.4.3.1 Single Exponential Smoothing. Based on NIST/SEMATECH e-Handbook of Statistical Methods
           http://www.itl.nist.gov/div898/handbook/pmc/section4/pmc431.htm

        Examples
        --------
        >>> mdspy.anomaly_detection.single_exponential_smoothing([0.15, 0.1, 0.15, 0.2, 0.1], 1., alpha=0.7, h=3)
        {'distance': 14.087944, 'prediction': 0.124555, 'range': 0.062141}
        >>> mdspy.anomaly_detection.single_exponential_smoothing([0.15, 0.1, 0.15, 0.2, 0.1], 1., alpha=1., h=3)
        {'distance': 12.727922, 'prediction': 0.1, 'range': 0.070711}
    '''

    assert isinstance(series, list)
    result = [np.nan, series[0]]
    for i in range(1, len(series)):
        result.append(alpha * series[i] + (1 - alpha) * result[i])
    assert len(result) == len(series) + 1
    mse = []
    assert len(series) - h >= 0
    for i in range(len(series) - h, len(series)):
        mse.append(abs(series[i] - result[i]) ** 2)
        # TODO why should we take the maximum between the squared error and 0.01 + where 0.01 comes from?
        # mse.append(max(abs(series[i] - result[i]) ** 2, 0.01))
    rmse = np.sqrt(np.mean(mse))
    output = {'distance': round(abs(sample - result[-1]) / rmse, 6), 'prediction': round(result[-1], 6),
              'range': round(rmse, 6)}
    return output


def double_exponential_smoothing(series, sample, alpha, beta, h=3):
    '''
        Predicts the value of the `sample` based on the exponential smoothing algorithm applied to the series of
        previous values.
        The prediction is also used to compute the error with the observation, normalized with the RMSE of previous
        errors.

        Parameters
        ----------
        series : array_like
            Input time series
        sample : float
            Sample (observation) to compare the prediction to
        alpha : float
            Weight (or smoothing factor) to assign to the previous sample
        beta : float
            Trend factor
        h : int
            Size of the history (default: 10 samples)

        Returns
        -------
        dict
            `distance`: the distance, expressed in standard deviations of the history values \n
            `prediction`: the prediction for the last value of the input time series \n
            `range`: the standard deviation of of the errors in the history

        Notes
        -----
        The system of two formulas used to compute the terms [1]_:

        .. math::
            S_t &= alpha * y_t + (1 - alpha) * (S_{t-1} + b_{t-1})

            b_t &= beta * (S_t - S_{t-1}) + (1 - beta) * b_{t-}1

        with `S` being the smoothed values, `b` the trend value, `y` the input series, alpha and beta the 2 required
        parameters and t the index.

        Initiation of S is often performed with S[1] = y[1]. Regarding b, 3 ways are suggested:

        .. math::
            b_1 &= y_2 - y_1

            b_1 &= 1/3 * ((y_2 - y_1) + (y_3 - y_2) + (y_4 - y_3))

            b_1 &= (y_n - y_1) / (n - 1)

        Forecasting is made with the following formula: :math:`F_{t+1} = S_t + b_t` [2]_

        References
        ----------
        .. [1] Section 6.4.3.3 Double Exponential Smoothing. Based on NIST/SEMATECH e-Handbook of Statistical Methods
           http://www.itl.nist.gov/div898/handbook/pmc/section4/pmc433.htm
        .. [2] Section 6.4.3.4 Forecasting with Double Exponential Smoothing (LASP). Based on NIST/SEMATECH e-Handbook
           of Statistical Methods http://www.itl.nist.gov/div898/handbook/pmc/section4/pmc434.htm

        Examples
        --------
        >>> y = [6.4, 5.6, 7.8, 8.8, 11., 11.6, 16.7, 15.3, 21.6]
        >>> sample = 22.4
        >>> mdspy.anomaly_detection.double_exponential_smoothing(y, sample, 0.3623, 1.0, h=3)
        {'distance': 0.228424, 'range': 2.849382, 'prediction': 23.050868}
    '''

    # result = [np.nan, series[0] + (series[1] - series[0])]
    result = [np.nan, series[0] + np.sum((series[1:4] - series[:3])) / 3]
    for n in range(1, len(series)):
        if n == 1:
            # level, trend = series[0], series[1] - series[0]
            level, trend = series[0], np.sum((series[1:4] - series[:3])) / 3
        if n >= len(series):
            value = result[-1]
        else:
            value = series[n]
        last_level, level = level, alpha * value + (1 - alpha) * (level + trend)
        trend = beta * (level - last_level) + (1 - beta) * trend
        result.append(level + trend)
    assert len(result) == len(series) + 1
    mse = []
    assert len(series) - h >= 0
    for i in range(len(series) - h, len(series)):
        mse.append(abs(series[i] - result[i]) ** 2)
    rmse = np.sqrt(np.mean(mse))
    output = {'distance': round(abs(sample - result[-1]) / rmse, 6), 'prediction': round(result[-1], 6),
              'range': round(rmse, 6)}
    return output



# coding=utf-8
import collections
import numpy as np
import pandas as pd
from scipy import stats
from scipy.spatial import distance
from sklearn.cluster import KMeans
import scipy.cluster.hierarchy as sch
import matplotlib.pylab as plt
import seaborn as sns


def clustering_hierarchical(df, cols, k=2, linkage_method='complete', fcluster_criterion='maxclust', color_threshold=2,
                            plot_dendrogram=False):
    """
    Performs a hierarchical clustering using scipy [1]_.

    Parameters
    ----------
    df : dataframe
        Input data frame
    cols : array_like
        List of columns to be considered from the data frame
    k : int
        Number of clusters
    linkage_method : str, optional
        Methods for calculating the distance between the clusters [2]_
    fcluster_criterion : str, optional
        The criterion to use in forming flat clusters [3]_
    color_threshold : float, optional
        Cut threshold to show different colors for different clusters [4]_
    plot_dendrogram : bool, optional
        When True, plots the dendrogram

    Returns
    -------
    clusters : array_like
        List with the cluster assigned to each line
    cluster_distribution: dict
        Dictionary with the count on the number of data points per cluster
    centroids : array_like
        List of all the centroids
    sum_intra_cluster_distances : float
        Sum of Squares of distances of the data points to their cluster center
    sum_inter_cluster_distances : float
        Sum of Squares of distances among all the centroids

    Notes
    -----
    Some references about the error metrics and how to better define the number of clusters [5]_ [6]_

    References
    ----------
    .. [1] Hierarchical clustering (scipy.cluster.hierarchy)
       https://docs.scipy.org/doc/scipy/reference/cluster.hierarchy.html

    .. [2] scipy.cluster.hierarchy.linkage
       https://docs.scipy.org/doc/scipy-0.18.1/reference/generated/scipy.cluster.hierarchy.linkage.html

    .. [3] scipy.cluster.hierarchy.fcluster
       https://docs.scipy.org/doc/scipy-0.18.1/reference/generated/scipy.cluster.hierarchy.fcluster.html

    .. [4] scipy.cluster.hierarchy.dendrogram
       https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.cluster.hierarchy.dendrogram.html

    .. [5] Data Mining Clustering
       http://wwwis.win.tue.nl/~tcalders/teaching/datamining09/slides/DM09-07-Clustering.pdf

    .. [6] Clustering Methods
       https://www.cs.swarthmore.edu/~meeden/cs63/s16/reading/Clustering.pdf

    Examples
    --------
    >>> import mdspy
    >>> import pandas as pd
    >>> dataset = [[1, 1], [1, 2], [1, 2], [5, 10], [2, 10]]
    >>> df_clustering = pd.DataFrame(dataset)
    >>> df_clustering.columns = ['col1', 'col2']
    >>> mdspy.analytics.clustering_hierarchical(df_clustering, ['col1', 'col2'], 2)
    {'clusters': [1, 1, 1, 2, 2], 'sum_intra_cluster_distances': 5.16666667,
    'sum_inter_cluster_distances': 75.694444444444443, 'centroids': [[1.0, 1.6666666666666667], [3.5, 10.0]],
    'cluster_distribution': {1: 3, 2: 2}}
    >>> mdspy.analytics.clustering_hierarchical(df_clustering, ['col1', 'col2'], 2, color_threshold=5, \
                                                plot_dendrogram=True)
    {'clusters': [1, 1, 1, 2, 2], 'sum_intra_cluster_distances':  5.16666667,
    'sum_inter_cluster_distances': 75.694444444444443, 'centroids': [[1.0, 1.6666666666666667], [3.5, 10.0]],
    'cluster_distribution': {1: 3, 2: 2}}

    """

    d = sch.distance.pdist(df[cols])
    z = sch.linkage(d, method=linkage_method)
    t = list(sch.fcluster(z, k, criterion=fcluster_criterion))
    cluster_distribution = dict(collections.Counter(t))

    df['clusters'] = t

    # calculate the centroids
    centroids = []
    centroids_reference = {}
    for c in list(set(t)):
        centroid = df[df['clusters'] == c][cols].mean(axis=0).values
        centroids_reference[c] = [centroid]
        centroids.append(list(centroid))

    # calculate the errors
    sum_intra_cluster_distances = 0
    for k, v in df.iterrows():
        sum_intra_cluster_distances += distance.cdist(centroids_reference[v['clusters']], [v[cols].values],
                                                      'euclidean') ** 2.0
    sum_intra_cluster_distances = sum_intra_cluster_distances[0][0]
    sum_inter_cluster_distances = (np.sum(distance.cdist(centroids, centroids, 'euclidean')) / 2.0) ** 2.0

    # plotting the dendrogram
    if plot_dendrogram:
        sns.set_style('whitegrid')
        p = sch.dendrogram(z, color_threshold=color_threshold)
        plt.show()

    return {'clusters': t, 'cluster_distribution': cluster_distribution, 'centroids': centroids,
            'sum_intra_cluster_distances': sum_intra_cluster_distances,
            'sum_inter_cluster_distances': sum_inter_cluster_distances}


def clustering_density(df, cols, k, n_jobs=1, n_init=10, random_state=None):
    """
    Performs a density clustering using K-Means. The implementation uses sklearn [1]_.

    Parameters
    ----------
    df : dataframe
        Input data frame
    cols : array_like
        List of columns to be considered from the data frame
    k : int
        Number of clusters
    n_jobs : int, optional
        The number of jobs to use for the computation. This works by computing each of the n_init runs in parallel.
        If -1 all CPUs are used.
    n_init : int, optional
        Number of time the k-means algorithm will be run with different centroid seeds. The final results will be the
        best output of n_init consecutive runs in terms of inertia.
    random_state : int, optional
        The generator used to initialize the centers. If an integer is given, it fixes the seed.

    Returns
    -------
    clusters : array_like
        List with the cluster assigned to each line
    cluster_distribution: dict
        Dictionary with the count on the number of data points per cluster
    centroids : array_like
        List of all the centroids
    sum_intra_cluster_distances : float
        Sum of Squares of distances of the data points to their cluster center
    sum_inter_cluster_distances : float
        Sum of Squares of distances among all the centroids

    Notes
    -----
    Some references about the error metrics and how to better define the number of clusters [2]_ [3]_

    References
    ----------
    .. [1] sklearn.cluster.KMeans
       http://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html

    .. [2] Data Mining Clustering
       http://wwwis.win.tue.nl/~tcalders/teaching/datamining09/slides/DM09-07-Clustering.pdf

    .. [3] Clustering Methods
       https://www.cs.swarthmore.edu/~meeden/cs63/s16/reading/Clustering.pdf


    Examples
    --------
    >>> import mdspy
    >>> import pandas as pd
    >>> mdspy.analytics.clustering_density(pd.DataFrame([[1, 1], [1, 2], [1, 2], [5, 10], [2, 10]]), 2)
    {'clusters': [0, 0, 0, 1, 1], 'sum_inter_cluster_distances': 8.7002554240921253,
    'centroids': [[1.0, 1.6666666666666665], [3.5, 10.0]], 'sum_intra_cluster_distances': 5.1666666666666661,
    'cluster_distribution': {0: 2, 1: 3}}
    >>> mdspy.analytics.clustering_density(pd.DataFrame([[1, 1], [1, 2], [1, 2], [5, 10], [2, 10]]), 2, 4, 50, 42))
    {'clusters': [1, 1, 1, 0, 0], 'sum_inter_cluster_distances': 8.7002554240921253,
    'centroids': [[3.5, 10.0], [1.0, 1.6666666666666665]], 'sum_intra_cluster_distances': 5.1666666666666661,
    'cluster_distribution': {0: 2, 1: 3}}
    """

    # set a random seed
    if random_state is None:
        random_state = np.random.random_integers(1000)
    kmeans = KMeans(k, n_jobs=n_jobs, n_init=n_init, random_state=random_state)
    kmeans.fit_transform(df[cols])
    T = list(kmeans.labels_)
    cluster_distribution = dict(collections.Counter(T))
    sum_intra_cluster_distances = kmeans.inertia_
    centroids = kmeans.cluster_centers_.tolist()
    sum_inter_cluster_distances = (np.sum(distance.cdist(centroids, centroids, 'euclidean')) / 2.0) ** 2.0

    return {'clusters': T, 'cluster_distribution': cluster_distribution, 'centroids': centroids,
            'sum_intra_cluster_distances': sum_intra_cluster_distances,
            'sum_inter_cluster_distances': sum_inter_cluster_distances}


def correlation(x, y):
    """
    Calculates a Pearson correlation coefficient and the p-value for testing non-correlation.

    Parameters
    ----------
    x : array_like
        Input sequences
    y : array_like
        Input sequences

    Returns
    -------
    pearson_correlation_coefficient : float
        Pearson's correlation coefficient
    interpretation : string
        Pearson's correlation coefficient interpretation

    Notes
    -----
    Following a table that explain the interpretation of the Pearson's correlation coefficient [1]_

    +---------------+-------------------------------+
    | Coefficient   | Strength of Relationship      |
    +===============+===============================+
    | [0.7, 1.0]    | Strong positive correlation   |
    +---------------+-------------------------------+
    | [0.5, 0.7[    | Moderate positive correlation |
    +---------------+-------------------------------+
    | [0.3, 0.5[    | Weak positive correlation     |
    +---------------+-------------------------------+
    | ]-0.3, 0.3[   | No correlation                |
    +---------------+-------------------------------+
    | [-0.3, -0.5[  | Weak negative correlation     |
    +---------------+-------------------------------+
    | [-0.5, -0.7[  | Moderate negative correlation |
    +---------------+-------------------------------+
    | [-0.7, -1.0]  | Strong negative correlation   |
    +---------------+-------------------------------+

    References
    ----------
    .. [1] Scatterplots and Correlation. Based on Chapter 4 of The Basic Practice of Statistics (6th ed.)
       https://www.westga.edu/academics/research/vrc/assets/docs/scatterplots_and_correlation_notes.pdf

    Examples
    --------
    >>> import mdspy
    >>> mdspy.analytics.correlation([1.0, 2.0, 2.0], [1.0, 2.0, 2.0])
    (1.0, 'Strong positive correlation')
    """

    r_row, p_value = stats.pearsonr(x, y)

    if r_row >= 0.7:
        interpretation = 'Strong positive correlation'
    elif r_row >= 0.5:
        interpretation = 'Moderate positive correlation'
    elif r_row >= 0.3:
        interpretation = 'Weak positive correlation'
    elif r_row <= -0.7:
        interpretation = 'Strong negative correlation'
    elif r_row <= -0.5:
        interpretation = 'Moderate negative correlation'
    elif r_row <= -0.3:
        interpretation = 'Weak negative correlation'
    else:
        interpretation = 'No correlation'

    return {'pearson_correlation_coefficient': r_row, 'interpretation': interpretation}


def best_fit_correlation(X):
    """
    Calculate the best fit of 78 different distributions to a dataset.

    Parameters
    ----------
    X : array_like
        Input sequences

    Returns
    -------
    distribution : str
        Distribution name. Distribution with highest score in the statistical test
    kolmogorov_smirnof_test : float
        The result of the Kolmogorov Smirnof test for the best fit distribution
    table_results : dataframe
        A table with the Kolmogorov Smirnof test results for 78 distributions

    Notes
    -----
    All the distributions calculated in this function are based on scipy [1]_

    References
    ----------
    .. [1] Statistical functions (scipy.stats)
       https://docs.scipy.org/doc/scipy/reference/stats.html

    Examples
    --------
    >>> import mdspy
    >>> x = [1, 1, 1, 2, 1, 2, 5, 10, 20]
    >>> mdspy.analytics.best_fit_correlation(x)
    {'distribution': 'logistic', 'table_results': DataFrame, 'kolmogorov_smirnof_test': 0.31093774290280862}

    """

    cdfs = [
        "norm",            #Normal (Gaussian)
        "alpha",           #Alpha
        "anglit",          #Anglit
        "arcsine",         #Arcsine
        "beta",            #Beta
        "betaprime",       #Beta Prime
        "bradford",        #Bradford
        "burr",            #Burr
        "cauchy",          #Cauchy
        "chi",             #Chi
        "chi2",            #Chi-squared
        "cosine",          #Cosine
        "dgamma",          #Double Gamma
        "dweibull",        #Double Weibull
        "expon",           #Exponential
        "exponweib",       #Exponentiated Weibull
        "exponpow",        #Exponential Power
        "fatiguelife",     #Fatigue Life (Birnbaum-Sanders)
        "foldcauchy",      #Folded Cauchy
        "f",               #F (Snecdor F)
        "fisk",            #Fisk
        "foldnorm",        #Folded Normal
        "frechet_r",       #Frechet Right Sided, Extreme Value Type II
        "frechet_l",       #Frechet Left Sided, Weibull_max
        "gamma",           #Gamma
        "gausshyper",      #Gauss Hypergeometric
        "genexpon",        #Generalized Exponential
        "genextreme",      #Generalized Extreme Value
        "gengamma",        #Generalized gamma
        "genlogistic",     #Generalized Logistic
        "genpareto",       #Generalized Pareto
        "genhalflogistic", #Generalized Half Logistic
        "gilbrat",         #Gilbrat
        "gompertz",        #Gompertz (Truncated Gumbel)
        "gumbel_l",        #Left Sided Gumbel, etc.
        "gumbel_r",        #Right Sided Gumbel
        "halfcauchy",      #Half Cauchy
        "halflogistic",    #Half Logistic
        "halfnorm",        #Half Normal
        "hypsecant",       #Hyperbolic Secant
        "invgamma",        #Inverse Gamma
        "invweibull",      #Inverse Weibull
        "johnsonsb",       #Johnson SB
        "johnsonsu",       #Johnson SU
        "laplace",         #Laplace
        "logistic",        #Logistic
        "loggamma",        #Log-Gamma
        "loglaplace",      #Log-Laplace (Log Double Exponential)
        "lognorm",         #Log-Normal
        "lomax",           #Lomax (Pareto of the second kind)
        "maxwell",         #Maxwell
        "mielke",          #Mielke's Beta-Kappa
        "nakagami",        #Nakagami
        "ncx2",            #Non-central chi-squared
        "nct",             #Non-central Student's T
        "pareto",          #Pareto
        "powerlaw",        #Power-function
        "powerlognorm",    #Power log normal
        "powernorm",       #Power normal
        "rdist",           #R distribution
        "reciprocal",      #Reciprocal
        "rayleigh",        #Rayleigh
        "rice",            #Rice
        "recipinvgauss",   #Reciprocal Inverse Gaussian
        "semicircular",    #Semicircular
        "t",               #Student's T
        "triang",          #Triangular
        "truncexpon",      #Truncated Exponential
        "truncnorm",       #Truncated Normal
        "tukeylambda",     #Tukey-Lambda
        "uniform",         #Uniform
        "vonmises",        #Von-Mises (Circular)
        "wald",            #Wald
        "weibull_min",     #Minimum Weibull (see Frechet)
        "weibull_max",     #Maximum Weibull (see Frechet)
        "wrapcauchy",      #Wrapped Cauchy
        "ksone",           #Kolmogorov-Smirnov one-sided (no stats)
        "kstwobign"]       #Kolmogorov-Smirnov two-sided test for Large

    fit = ['', 0.0]
    df_fit = []
    for cdf in cdfs:
        #fit our data set against every probability distribution
        parameters = eval("stats."+cdf+".fit(X)")
        #Applying the Kolmogorov-Smirnof one sided test
        D, p = stats.kstest(X, cdf, args=parameters)
        df_fit.append([cdf, round(p, 6)])
        if p > fit[1]:
            fit = [cdf, p]
    df_fit = pd.DataFrame(df_fit)
    df_fit.columns = ['Distribution', 'Kolmogorov Smirnof Test']
    df_fit = df_fit.sort_values('Kolmogorov Smirnof Test', ascending=False)

    return {'distribution': fit[0], 'kolmogorov_smirnof_test': fit[1], 'table_results': df_fit}



# coding=utf-8
import pandas as pd
import datetime
from scipy import stats
import numpy as np


def ttt_ratio(ts, target):
    """
    Performs the time to target using the sum of deltas. This approach is similar to calculate the ratio considering
    the first and last point.

    TODO: - This implementation is not considering "recharged"

    TODO: - This implementation doesn't remove outliers

    TODO: - This implementation doesn't trait corner cases (example: slope == 0)

    Parameters
    ----------
    ts : Series
        pandas Series used in the training process
    target : number
        The target

    Returns
    -------
    time_to_target : float
        The number of days predicted to reach the target

    References
    ----------

    Examples
    --------
    >>> ts_test = pd.Series(
    >>>     index=pd.to_datetime(['2017-01-01', '2017-01-02', '2017-01-03', '2017-01-04', '2017-01-05']),
    >>>     data=[10, 5, 10, 5, 4])
    >>> target = 0
    >>> mdspy.ttt.ttt_ratio(ts_test, target)
    2.666666666
    """
    diffs = sum(ts.diff(1).fillna(0.0))
    delta = ts.tail(1).index[0] - ts.head(1).index[0]
    delta_days = delta.days
    last_value = ts.tail(1).values[0]
    ratio = abs(diffs / delta_days)
    return (last_value - target) / ratio


def ttt_ratio_ma(ts, target, w):
    """
    Initially smooth the time series applying a moving average. After, performs the time to target using the sum of
    deltas. This approach is similar to calculate the ratio considering the first and last point.
    The same as ttt_ratio but with a moving average smooth.

    TODO: - This implementation is not considering "recharged"

    TODO: - This implementation doesn't remove outliers

    TODO: - This implementation doesn't trait corner cases (example: slope == 0)

    Parameters
    ----------
    ts : Series
        pandas Series used in the training process
    target : number
        The target
    w : int
        Size of window to smooth the time series

    Returns
    -------
    time_to_target : float
        The number of days predicted to reach the target

    References
    ----------

    Examples
    --------
    >>> ts_test = pd.Series(
    >>>     index=pd.to_datetime(['2017-01-01', '2017-01-02', '2017-01-03', '2017-01-04', '2017-01-05']),
    >>>     data=[10, 5, 10, 5, 4])
    >>> target = 0
    >>> window = 2
    >>> mdspy.ttt.ttt_ratio_ma(ts_test, target, window)
    5.3333333
    """
    diffs = sum(ts.rolling(window=w).mean().diff(1).fillna(0.0))
    delta = ts.tail(1).index[0] - ts.head(1).index[0]
    delta_days = delta.days
    initial = ts.tail(1).values[0]
    ratio = abs(diffs / delta_days)
    return (initial - target) / ratio


def ttt_linear_regression(ts, target):
    """
    Performs a linear regression and use the slope to calculate the time to target.

    TODO: - This implementation is not considering "recharged"

    TODO: - This implementation doesn't remove outliers

    TODO: - This implementation doesn't trait corner cases (example: slope == 0)

    Parameters
    ----------
    ts : Series
        pandas Series used in the training process
    target : number
        The target

    Returns
    -------
    time_to_target : float
        The number of days predicted to reach the target

    References
    ----------

    Examples
    --------
    >>> ts_test = pd.Series(
    >>>     index=pd.to_datetime(['2017-01-01', '2017-01-02', '2017-01-03', '2017-01-04', '2017-01-05']),
    >>>     data=[10, 5, 10, 5, 4])
    >>> target = 0
    >>> mdspy.ttt.ttt_linear_regression(ts_test, target)
    3.3333333
    """
    epoch = ts.index.astype(np.int64) / 1000000000
    slope, intercept, r_value, p_value, std_err = stats.linregress(epoch, ts.values)
    prediction = ((target - ts.tail(1).values[0])/slope) / (60.0 * 60.0 * 24.0)
    return prediction



#!/usr/bin/env python
from setuptools import setup, find_packages

with open('README.rst') as fread:
    readme = fread.read()

with open('requirements.txt') as f:
    requirements = [line.strip() for line in f.readlines()]
VERSION = "1.0{tag}"
try:
    with open('build_version.txt') as f:
        tag = f.readline().strip()
        version = VERSION.format(tag="."+tag)
except IOError:
    version = VERSION.format(tag="-dev")
setup(
    name="mdspy",
    version=version,
    description="Mnubo Data Science Library",
    author="mnubo, inc.",
    author_email="support@mnubo.com",
    #url="https://github.com/mnubo/smartobjects-python-client",
    #packages=["smartobjects", "smartobjects.ingestion", "smartobjects.restitution"],
    install_requires=requirements,
    long_description=readme,
    packages=find_packages(exclude=('test', 'doc')),
    keywords=['mnubo', 'iot', 'smartobjects'],
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
)